"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

#!/usr/bin/env -S python -W ignore

import os
import sys

license_header = """'''
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''
""".splitlines()

_hint = sum(len(s) for s in license_header) + 7
_fix = len(sys.argv) == 2 and sys.argv[1] == "--fix"


def _process_file(file_path):
    compliant, fixable = _check_file_license(file_path)
    if compliant:
        return

    if not _fix or not fixable:
        print("ERROR", file_path)
        sys.exit(1)

    print("Updating file", file_path)
    with open(file_path, "r+") as f:
        content = f.read()
        f.seek(0, 0)
        for line in license_header:
            f.write(line + "\n")
        f.write(content)


def _check_file_license(file_path):
    with open(file_path, "r") as f:
        lines = f.readlines(_hint)
        if len(lines) == 0:
            return True, False

        fixable = lines[0][0] != "#"
        if len(lines) < len(license_header):
            return False, fixable

        for i in range(len(license_header)):
            if lines[i].strip() != license_header[i]:
                return False, fixable
    return True, False


if __name__ == "__main__":
    for root, subdirs, files in os.walk("."):
        for filename in files:
            if not filename.endswith(".py"):
                continue
            file_path = os.path.join(root, filename)
            _process_file(file_path)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from copy import deepcopy
from typing import Callable

import hcs_core.util.duration as duration
from hcs_core.plan import PluginException, actions
from hcs_core.util import exit


def deploy(data: dict, state: dict, save_state: Callable) -> dict:
    text = data.get("text")
    error = data.get("error")
    delay = data.get("delay")
    timeout = data.get("timeout")

    # import json
    # print(json.dumps(data, indent=4))

    a = data.get("a", 0)
    b = data.get("b", 0)
    error_before_save = data.get("error_before_save")
    delay_seconds = duration.to_seconds(delay)
    if delay_seconds:
        exit.sleep(delay_seconds)

    ret = deepcopy(data)
    ret["outputText"] = text
    ret["n"] = a + b

    if not error_before_save:
        save_state(ret)

    if error:
        raise PluginException(error)
    if timeout:
        raise TimeoutError("Simulated timeout")

    return ret


def refresh(data: dict, state: dict) -> dict:
    return state


def decide(data: dict, state: dict):
    return actions.skip


def destroy(data: dict, state: dict, force: bool) -> dict:
    return {}


def eta(action: str, data: dict, state: dict):
    if action == actions.create:
        delay = data.get("delay")
        delay_seconds = duration.to_seconds(delay)
        return f"{delay_seconds}s"
    if action == actions.delete:
        return "1s"



import json

import yaml

from ..ctxp import CtxpException
from .cli_options import get_org_id


def get_payload_with_defaults(file_arg, org: str):
    with file_arg:
        text = file_arg.read()

    try:
        payload = json.loads(text)
    except Exception as e1:
        try:
            payload = yaml.safe_load(text)
        except Exception:
            raise CtxpException("Invalid payload: " + str(e1))

    # Override org id, if specified explicitly on args.
    # Otherwise override using default, if not in payload.
    if org:
        payload["orgId"] = org
    else:
        if not payload.get("orgId"):
            payload["orgId"] = get_org_id()
    return payload



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import ABC, abstractmethod
from typing import Callable

from .actions import actions


class BaseProvider(ABC):
    def __init__(self, data: dict, state: dict, fn_save_state: Callable):
        self.data = data
        self.state = state
        self._save_state = fn_save_state

    def save(self, data: dict):
        self._save_state(data)

    def decide(self) -> str:
        return actions.create

    @abstractmethod
    def refresh(self) -> dict:
        pass

    @abstractmethod
    def create(self) -> dict:
        pass

    @abstractmethod
    def delete(self) -> dict:
        pass

    def eta_create(self):
        return "1m"

    def eta_delete(self):
        return "1m"



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from os import path
from pathlib import Path

import click

from . import cli_processor, config, profile, state

user_home = str(Path.home())


_initialized_app_name = None


def init(app_name: str, store_path: str = user_home, config_path: str = "./config"):
    global _initialized_app_name
    if _initialized_app_name == app_name:
        return
    if _initialized_app_name is not None:
        raise ValueError(f"App {app_name} already initialized with {_initialized_app_name}")
    _initialized_app_name = app_name
    try:
        real_store_path = path.join(store_path, "." + app_name)
        state.init(real_store_path, ".state")
        profile.init(real_store_path)
        config.init(config_path)

    except Exception as e:
        # critical errors, must print stack trace.
        if _need_stack_trace(e):
            raise e
        else:
            # Other errors, no stack.
            from .util import panic

            panic(e)


# init default with env if configured
_app_name = os.environ.get("CTXP_APP_NAME")
if _app_name:
    init(_app_name)


def app_name():
    dir_name = path.dirname(state._file._path)
    name = dir_name[dir_name.rindex("/") + 1 :]
    if name.startswith("."):
        return name[1:]
    raise ValueError("Unable to determine app name: " + dir_name)


def init_cli(main_cli: click.Group, commands_dir: str = "./cmds"):
    try:
        return cli_processor.init(main_cli, commands_dir)
    except Exception as e:
        if _need_stack_trace(e):
            raise e
        else:
            from .util import panic

            panic(e)


def _need_stack_trace(e) -> bool:
    program_errors = [
        LookupError,
        TypeError,
        ValueError,
        ArithmeticError,
        NameError,
        SyntaxError,
        KeyError,
        AttributeError,
    ]
    for t in program_errors:
        if isinstance(e, t):
            return True
    return False



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os
import sys
import time
from typing import Callable, Iterator

from hcs_core.ctxp import CtxpException, panic, profile
from hcs_core.sglib import hcs_client
from hcs_core.sglib.ez_client import EzClient
from hcs_core.util import duration, exit
from hcs_core.util.query_util import PageRequest, with_query

_caches = {}


def hdc_service_client(service_name: str) -> EzClient:
    instance = _caches.get(service_name)
    if not instance:

        def _get_url():  # make it deferred so no need to initialize profile
            url = _get_hcs_url_considering_env_override(service_name)
            if not url.endswith("/"):
                url += "/"
            url += service_name
            return url

        instance = hcs_client(_get_url)
        _caches[service_name] = instance
    return instance


def _get_hcs_url_considering_env_override(service_name: str):
    env_name_alt_url = f"hcs_{service_name}_url"
    url = os.environ.get(env_name_alt_url)
    if url:
        return url
    return profile.current().hcs.url


def _get_region_url(region_name: str):
    regions = profile.current().hcs.regions
    if not region_name:
        return regions[0].url
    for r in regions:
        if r.name.lower() == region_name.lower():
            return r.url
    names = []
    for r in regions:
        names.append(r.name)
    panic(f"Region not found: {region_name}. Available regions: {names}")


def regional_service_client(region_name: str, service_name: str):
    # 'https://dev1b-westus2-cp103a.azcp.horizon.vmware.com/vmhub'
    url = _get_region_url(region_name)
    if not url:
        panic("Missing profile property: hcs.regions")
    if not url.endswith("/"):
        url += "/"
    url += service_name
    return hcs_client(url)


def _with_org_id(url: str, org_id: str):
    if org_id:
        if url.find("?") < 0:
            url += "?"
        url += "org_id=" + org_id
    return url


class default_crud:
    def __init__(self, client, base_context: str, resource_type_name: str):
        self._client_impl = client
        self._base_context = base_context
        self._resource_type_name = resource_type_name

    def _client(self):
        if callable(self._client_impl):
            self._client_impl = self._client_impl()
        elif isinstance(self._client_impl, str):
            self._client_impl = hdc_service_client(self._client_impl)
        else:
            pass
        if isinstance(self._client_impl, EzClient):
            return self._client_impl
        raise CtxpException(f"Invalid client implementation: {self._client_impl}")

    def get(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client().get(url)

    def list(self, org_id: str, fn_filter: Callable = None, **kwargs) -> list:
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id

        def _get_page(query_string):
            url = self._base_context + "?" + query_string
            # print(url)
            return self._client().get(url)

        return PageRequest(_get_page, fn_filter, **kwargs).get()

    def items(self, org_id: str, fn_filter: Callable = None, **kwargs) -> Iterator:
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id

        def _get_page(query_string):
            url = self._base_context + "?" + query_string
            return self._client().get(url)

        return PageRequest(_get_page, fn_filter, **kwargs).items()

    def create(self, payload: dict, headers: dict = None, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        # print(url)
        # import json
        # print(json.dumps(payload, indent=4))
        if isinstance(payload, str):
            return self._client().post(url, text=payload, headers=headers)
        if isinstance(payload, dict):
            return self._client().post(url, json=payload, headers=headers)
        return self._client().post(url, json=payload, headers=headers)

    def upload(self, files, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        return self._client().post(url, files=files)

    def delete(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client().delete(url)

    def wait_for_deleted(self, id: str, org_id: str, timeout: str, fn_is_error: Callable = None):
        name = self._resource_type_name + "/" + id
        fn_get = lambda: self.get(id, org_id)
        return wait_for_res_deleted(name, fn_get, timeout=timeout, fn_is_error=fn_is_error)

    def update(self, id: str, org_id: str, data: dict, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}")
        return self._client().patch(url, data)


def _parse_timeout(timeout: str):
    if isinstance(timeout, int):
        return timeout
    if isinstance(timeout, str):
        return duration.to_seconds(timeout)

    raise CtxpException(f"Invalid timout. Type={type(timeout).__name__}, value={timeout}")


def wait_for_res_deleted(
    resource_name: str,
    fn_get: Callable,
    timeout: str,
    polling_interval_seconds: int = 10,
    fn_is_error: Callable = None,
):
    timeout_seconds = _parse_timeout(timeout)
    start = time.time()
    while True:
        t = fn_get()
        if t is None:
            return
        if fn_is_error:
            if fn_is_error(t):
                msg = f"Failed deleting resource '{resource_name}', resource in Error state."
                raise CtxpException(msg)

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            msg = f"Timeout waiting for resource '{resource_name}' to be deleted."
            raise TimeoutError(msg)
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds
        exit.sleep(sleep_seconds)


def wait_for_res_status(
    resource_name: str,
    fn_get: Callable,
    get_status: Callable,
    status_map: dict = None,
    is_ready: Callable = None,
    is_error: Callable = None,
    is_transition: Callable = None,
    timeout: str = "10m",
    polling_interval: str = "20s",
    not_found_as_success: bool = False,
):
    timeout_seconds = _parse_timeout(timeout)
    polling_interval_seconds = _parse_timeout(polling_interval)
    if polling_interval_seconds < 3:
        polling_interval_seconds = 3
    start = time.time()
    prefix = f"Error waiting for resource {resource_name}: "

    if isinstance(get_status, str):
        field_name = get_status
        get_status = lambda t: t[field_name]
    if status_map:
        if isinstance(status_map["ready"], str):
            status_map["ready"] = [status_map["ready"]]
        if isinstance(status_map["transition"], str):
            status_map["transition"] = [status_map["transition"]]
        if isinstance(status_map["error"], str):
            status_map["error"] = [status_map["error"]]
        if is_ready:
            raise CtxpException("Can not specify is_ready when status_map is provided.")
        if is_error:
            raise CtxpException("Can not specify is_error when status_map is provided.")
        if is_transition:
            raise CtxpException("Can not specify is_transition when status_map is provided.")
        is_ready = lambda s: s in status_map["ready"]
        is_error = lambda s: s in status_map["error"]
        is_transition = lambda s: s in status_map["transition"]
    else:
        if not is_ready:
            raise CtxpException("Either status_map or is_ready must be specified.")
        if not is_error:
            raise CtxpException("Either status_map or is_error must be specified.")
        if not is_transition:
            raise CtxpException("Either status_map or is_transition must be specified.")

    while True:
        t = fn_get()
        if t is None:
            if not_found_as_success:
                return
            raise CtxpException(prefix + "Not found.")
        status = get_status(t)
        # print(f"DDD - status {status}")
        if is_error(status):
            msg = prefix + f"Status error. Actual={status}"
            if status_map:
                msg += f", expected={status_map['ready']}"
            print("-- DUMP START --", file=sys.stderr)
            print(json.dumps(t, indent=4), file=sys.stderr)
            print("-- DUMP END --", file=sys.stderr)
            raise CtxpException(msg)
        if is_ready(status):
            return t
        if not is_transition(status):
            raise CtxpException(
                prefix + f"Unexpected status: {status}. If this is a transition, add it to status_map['transition']."
            )

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            raise TimeoutError(prefix + "Timeout.")
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds

        # print(f"DDD sleeping {sleep_seconds}")
        exit.sleep(sleep_seconds)



import json
import re
from io import TextIOWrapper
from os import chmod, path
from typing import Any, Tuple

from .util import CtxpException


def load_data(file_name: str, class_type: str):
    data = load_data_file(file_name)
    if data is None:
        return
    return strict_dict_to_class(data, class_type)


def load_data_file(file, default=None, format="auto"):

    if isinstance(file, TextIOWrapper):
        text = file.read(-1)
        text = text.strip()

        if format == "json":
            data = json.loads(text)
        elif format == "yaml" or format == "yml":
            import yaml

            data = yaml.safe_load(text)
        else:
            try:
                import yaml

                data = yaml.safe_load(text)
            except Exception:
                data = json.loads(text)
        return default if data is None else data
    else:
        if not path.exists(file) or not path.isfile(file):
            return default
        with open(file, encoding="utf-8") as f:
            text = f.read()

        _, ext = path.splitext(file)
        if ext == ".json" or format == "json":
            if format != "auto" and format != "json":
                raise CtxpException(f"File extension does not match specified format. File={file}, format={format}")
            data = json.loads(text)
            return default if data is None else data  # handle empty file
        if ext == ".yaml" or ext == ".yml" or format == "yml" or format == "yaml":
            if format != "auto" and format != "yaml" and format != "yml":
                raise CtxpException(f"File extension does not match specified format. File={file}, format={format}")
            import yaml

            data = yaml.safe_load(text)
            return default if data is None else data  # handle empty file
    return text


def save_data_file(data, file_name: str, format: str = "yaml", file_mod: int = 0):
    with open(file_name, "w") as file:
        if format == "yaml":
            import yaml

            # TODO
            # yaml.safe_dump(data, file, sort_keys=False)
            yaml.safe_dump(json.loads(json.dumps(data, default=vars)), file, sort_keys=False)
        elif format == "json":
            json.dump(data, file, indent=4, default=vars)
        else:
            raise CtxpException("Invalid format: " + format)

    if file_mod:
        chmod(file_name, file_mod)


def strict_dict_to_class(data: dict, class_type):
    actual_keys = set(data.keys())
    declared_keys = set(class_type.__annotations__.keys())
    unexpected_fields = actual_keys - declared_keys
    if unexpected_fields:
        raise ValueError(f"Unexpected fields: {unexpected_fields} while deserializing class {class_type.__name__}")

    mandatory_keys = set()
    for k in declared_keys:
        if not hasattr(class_type, k):
            mandatory_keys.add(k)
    missing_fields = mandatory_keys - actual_keys
    if missing_fields:
        raise ValueError(f"Missing fields: {class_type.__name__}.{missing_fields}")

    inst = class_type()

    for field_name, field_type in class_type.__annotations__.items():
        value = data.get(field_name)
        if isinstance(value, field_type):
            setattr(inst, field_name, value)
            continue
        if isinstance(value, dict):
            value = strict_dict_to_class(value, field_type)
            setattr(inst, field_name, value)
            continue
        raise ValueError(
            f"Field '{class_type.__name__}.{field_name}' has an incorrect type. Declared: {field_type}, actual: {type(value)}"
        )
    return inst


def deep_update_object_value(obj, fn_change, _current_path: str = ""):
    if isinstance(obj, list):
        for i in range(len(obj)):
            obj[i] = deep_update_object_value(obj[i], fn_change, _current_path + f"[{i}]")
    elif isinstance(obj, dict):
        for k, v in obj.items():
            item_path = _current_path + "." + k if _current_path else k
            obj[k] = deep_update_object_value(v, fn_change, item_path)
    else:
        obj = fn_change(_current_path, obj)
    return obj


def deep_apply_default(to_obj: dict, from_obj: dict) -> bool:
    """If a property in to_obj is empty, use the value from from_obj if the same property is not empty"""
    if not from_obj:
        return
    changed = False
    for k, v in to_obj.items():
        v1 = from_obj.get(k)
        if not v1:
            continue
        if not v:
            to_obj[k] = v1
            changed = True
            continue
        if isinstance(v, dict):
            if deep_apply_default(v, v1):
                changed = True
    return changed


def deep_get_attr(obj: dict, path: str, raise_on_not_found: bool = True):
    parts = path.split(".")
    for k in parts:
        try:
            obj = _get_obj_attr(obj, k)
        except KeyError:
            if raise_on_not_found:
                raise CtxpException("Property path not found: " + path)
            else:
                return
    return obj


def deep_set_attr(obj: dict, path: str, value, raise_on_not_found: bool = True):
    parts = path.split(".")
    if len(parts) == 1:
        obj[path] = value
        return obj

    k = None
    try:
        for i in range(len(parts)):
            k = parts[i]
            try:
                sub_obj = _get_obj_attr(obj, k)
            except KeyError:
                if raise_on_not_found:
                    raise
                sub_obj = {}
                obj[k] = sub_obj

            obj = sub_obj

            if i == len(parts) - 2:
                # found the one before the leaf.
                _set_obj_attr(obj, parts[i + 1], value)
                break
        return obj
    except (KeyError, TypeError, IndexError) as e:
        raise CtxpException(f"Property path error: {path}. Cause={e}, current={k}, i={i}")


def _get_obj_attr(o, k):
    name, array_index = _parse_array_property_name(k)

    if isinstance(o, dict):
        ret = o[name]
    else:
        ret = getattr(o, name)

    if array_index is not None:
        ret = ret[array_index]

    return ret


def _set_obj_attr(o, k, v):
    name, array_index = _parse_array_property_name(k)
    if isinstance(o, dict):
        if array_index is None:
            o[name] = v
        else:
            o[name][array_index] = v
    else:
        if array_index is None:
            setattr(o, name, v)
        else:
            array_elem = getattr(o, name)
            array_elem[array_index] = v


_array_index_matcher = re.compile("(.+)\\[(\\d+)\\]")


def _parse_array_property_name(k: str) -> Tuple[str, int]:
    m = _array_index_matcher.match(k)
    if m:
        return m.group(1), int(m.group(2))
    return k, None


def deep_iterate(obj, fn_on_value):
    if isinstance(obj, list) or isinstance(obj, set):
        for v in obj:
            deep_iterate(v, fn_on_value)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            deep_iterate(v, fn_on_value)
    else:
        fn_on_value(obj)


def deep_find_variables(obj):
    collector = set()

    def fn_on_value(v):
        if not v or not isinstance(v, str):
            return
        matches = re.findall(_pattern_var_only, v)
        if matches:
            m2 = _pattern_var_list.match(v)
            if m2:
                collector.add(m2.group(2))
            else:
                for i in matches:
                    collector.add(i)

    deep_iterate(obj, fn_on_value)
    return collector


def process_variables(obj: dict, fn_get_var=None, use_env: bool = True):
    if fn_get_var is None:

        def _fn_get_var(name):
            try:
                return deep_get_attr(obj, name), True
            except Exception:
                return None, False

        fn_get_var = _fn_get_var

    if use_env:
        import os

        prev_fn_get_var = fn_get_var

        def _fn_get_var_from_env(name: str):
            if not name.startswith("env."):
                return prev_fn_get_var(name)
            v = None
            actual_name = name[4:]
            v = os.environ.get(actual_name, None)
            if v:
                return v, True
            v, found = prev_fn_get_var(name)
            if found:
                return v
            raise CtxpException(f"Environment variable '{actual_name}' is used in template, but not found. ")

        fn_get_var = _fn_get_var_from_env

    total_changed = {}
    while True:
        ret = _process_variables_impl(obj, fn_get_var)
        total_changed.update(ret["changed"])
        if not ret["changed"]:
            return {
                "data": ret["data"],
                "changed": total_changed,
                "pending": ret["pending"],
            }


_pattern_var = re.compile(".*\\$\\{(.+?)\\}.*")
_pattern_var_only = re.compile("\\$\\{(.+?)\\}")
_pattern_var_list = re.compile("\\$\\{\\s*\\[\\s*for\\s+(.+?)\\s+in\\s+(.+?)\\s*\\:\\s*(.+)\\s*]\\s*\\}")


def _process_variables_impl(obj: dict, fn_get_var=None):
    changed = {}
    pending = {}

    def fn_change(path, v):
        resolved, pending_var = resolve_expression(v, fn_get_var, path)
        if pending_var:
            pending[path] = pending_var
        elif resolved != v:
            changed[path] = v
        return resolved

    data = deep_update_object_value(obj, fn_change)
    return {"changed": changed, "pending": pending, "data": data}


def resolve_expression(expr, fn_get_value, referencing_attr_path) -> Tuple[Any, str]:
    """Try resolving expression. Returned updated value and None, or value and pending var name"""

    if not isinstance(expr, str):
        return expr, None
    # get variable names from expr
    m1 = _pattern_var.match(expr)
    if not m1:
        return expr, None
    m2 = _pattern_var_list.match(expr)
    if m2:
        # expression match
        tmp_var_name = m2.group(1)
        src_var_name = m2.group(2)
        mapped_value = m2.group(3)

        target_value, found = fn_get_value(src_var_name)
        if not found:
            return expr, src_var_name
        if not isinstance(target_value, list):
            raise CtxpException(
                f"Invalid variable value for expression. Expect list, actual {type(target_value).__name__}. attr_path={referencing_attr_path}, src_var_name={src_var_name}"
            )
        if not mapped_value.startswith(tmp_var_name + "."):
            raise CtxpException(
                f"Unsupported expression. attr_path={referencing_attr_path}, src_var_name={src_var_name}"
            )
        new_attr_path = mapped_value[len(tmp_var_name) + 1 :]
        ret = []
        for i in target_value:
            if i is None:
                # raise CtxpException(f"Invalid value encountered during processing array expansion expression. The value in the referenced array is None. This is normally a problem of the data, not the caller. attr_path={referencing_attr_path}, src_var_name={src_var_name}")
                # There are valid cases that some of such items are null. Raising exception
                # makes the handling hard. Pass-on null in such scenario, so enable the downstream
                # components to decide what to do with the null case.
                item = None
            else:
                item = deep_get_attr(i, new_attr_path)
            ret.append(item)
        return ret, None  # replace the entire value using the new value.
    else:
        var_name = m1.group(1)
        # string replacement
        actual_value, found = fn_get_value(var_name)
        if not found:
            return expr, var_name

        # if the replacement is the entire value, not part of a string:
        if len(expr) == len(var_name) + 3:
            return actual_value, None

        # the var is part of a string. Do string replacement
        if isinstance(actual_value, str):
            return expr.replace("${" + var_name + "}", actual_value), None
        if isinstance(actual_value, int):
            return expr.replace("${" + var_name + "}", str(actual_value)), None
        # replacement is an object.
        raise CtxpException(
            f"Invalid variable: can not replace variable in string with non-string. attr_path={referencing_attr_path}, var_name={var_name}, replacement={actual_value}, expr={expr}"
        )


def to_json(o) -> str:
    class SetEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, set):
                return list(obj)
            return json.JSONEncoder.default(self, obj)

    return json.dumps(o, cls=SetEncoder, indent=4)


def get_common_items(iter1, iter2):
    set1 = set(iter1)
    set2 = set(iter2)
    common_items = set1.intersection(set2)
    return common_items


def _evaluate(value, smart_search):
    # If we found an exact match, return 2 immediately
    if value == smart_search:
        return 2
    # Check if the smart_search is a substring of value
    elif smart_search in value:
        return 1
    return 0


def deep_search(target, smart_search):
    """
    Recursively searches through a dictionary (and any nested dictionaries or lists)
    to find if any string value contains the 'smart_search' substring.

    Parameters:
    - target (dict): The dictionary to search through.
    - smart_search (str): The substring to search for within the dictionary values.

    Returns:
    - int:
        - 0: if 'smart_search' is not found
        - 1: if 'smart_search' appears as substring
        - 2: if 'smart_search' appears as exact match of a value.
    """

    def _dfs_with_prune(node):
        match_found = 0
        if isinstance(node, str):
            return _evaluate(node, smart_search)
        elif isinstance(node, dict):
            for value in node.values():
                score = _dfs_with_prune(value)
                if score == 2:  # Exact match found, prune search
                    return 2
                elif score == 1:
                    match_found = 1
        elif isinstance(node, list):
            for item in node:
                score = _dfs_with_prune(item)
                if score == 2:  # Exact match found, prune search
                    return 2
                elif score == 1:
                    # Found a substring match, but continue to search for exact match
                    match_found = 1
        return match_found

    return _dfs_with_prune(target)



import logging
import threading
from copy import deepcopy
from dataclasses import dataclass
from time import sleep, time

import schedule

from . import duration

log = logging.getLogger(__name__)


class TaskRef:
    def __init__(self, ref1, ref2=None):
        self.ref1 = ref1
        self.ref2 = ref2

    def cancel(self):
        if self.ref1:
            fn = list(self.ref1.job_func.args)[0]
            log.debug("Cancel job %s: %s" % (fn.__name__, self.ref1))
            schedule.cancel_job(self.ref1)
            self.ref1 = None
        if self.ref2:
            fn2 = list(self.ref2.job_func.args)[0]
            log.debug("Cancel job %s: %s" % (fn2.__name__, self.ref2))
            schedule.cancel_job(self.ref2)
            self.ref2 = None


@dataclass
class JobStatistics:
    last_start: int = 0
    last_end: int = 0
    history_total: int = 0
    last_cycle_total: int = 0

    def start(self):
        self.last_start = int(time())
        self.last_cycle_total += 1
        self.history_total += 1

    def end(self):
        self.last_end = int(time())

    def reset_cycle(self):
        self.last_cycle_total = 0


@dataclass
class Statistics:
    run_once_job: JobStatistics = JobStatistics()
    recurring_job: JobStatistics = JobStatistics()
    size: int = 0

    def reset_cycle(self):
        self.run_once_job.reset_cycle()
        self.recurring_job.reset_cycle()


_g_worker_thread: threading.Event = None
_g_flag_stop_daemon: threading.Thread = None
_g_flag_running: bool = True
_g_statistics = Statistics()


def _task_wrapper_repeat(fn_impl, kwargs):
    _g_statistics.recurring_job.start()
    fn_impl(**kwargs)
    _g_statistics.recurring_job.end()


def _task_wrapper_once(fn_impl, kwargs):
    _g_statistics.run_once_job.start()
    fn_impl(**kwargs)
    _g_statistics.run_once_job.end()
    return schedule.CancelJob


def submit(fn_task: callable, initial_delay: str = None, repeat_interval: str = None, **kwargs):
    if initial_delay:
        initial_delay_seconds = duration.to_seconds(initial_delay)
    else:
        initial_delay_seconds = 1

    if repeat_interval:
        # Have an initial run
        # TODO: this is not correct, but better than no lib-provided initial delay.
        job1 = schedule.every(initial_delay_seconds).seconds.do(_task_wrapper_once, fn_task, kwargs)
        log.debug(f"Register initial run {fn_task.__name__} {job1}")

        # TODO: identify a way to use the scheduler with initial delay, with cancellation in mind.
        seconds = duration.to_seconds(repeat_interval)
        job2 = schedule.every(seconds).seconds.do(_task_wrapper_repeat, fn_task, kwargs)
        log.debug(f"Register scheduled job {fn_task.__name__} at interval {repeat_interval} {job2}")
    else:
        job1 = schedule.every(initial_delay_seconds).seconds.do(_task_wrapper_once, fn_task, kwargs)
        log.debug(f"Register one-shot job {fn_task.__name__} {job1}")
        job2 = None
    return TaskRef(job1, job2)


def statistics(reset_cycle: bool = False):
    _g_statistics.size = len(schedule.get_jobs())
    ret = deepcopy(_g_statistics)
    if reset_cycle:
        _g_statistics.reset_cycle()
    return ret


def _daemon_worker():
    log.info("task scheduler daemon thread start")
    global _g_flag_stop_daemon
    while not _g_flag_stop_daemon.is_set():
        if _g_flag_running:
            schedule.run_pending()
        sleep(1)
    log.info("task scheduler daemon thread exit")


def start_daemon(paused: bool = False):
    global _g_flag_stop_daemon
    global _g_worker_thread
    if _g_worker_thread:
        raise Exception("Already started")

    if paused:
        global _g_flag_running
        _g_flag_running = False

    _g_flag_stop_daemon = threading.Event()
    _g_worker_thread = threading.Thread(target=_daemon_worker, daemon=True, name="task-schd")
    _g_worker_thread.start()


def pause():
    global _g_flag_running
    if _g_flag_running:
        _g_flag_running = False
        log.info("task scheduler daemon paused")


def resume():
    global _g_flag_running
    if not _g_flag_running:
        _g_flag_running = True
        log.info("task scheduler daemon resumed")


def stop_daemon():
    global _g_flag_stop_daemon
    global _g_worker_thread
    if _g_worker_thread:
        _g_flag_stop_daemon.set()
        _g_worker_thread.join()
        _g_worker_thread = None


if __name__ == "__main__":

    def job1(a, b):
        log.info(f"job1 {a} {b}")

    # logutil.setup()
    start_daemon()
    j1 = submit(fn_task=job1, initial_delay="PT3S", repeat_interval="PT5S", a="aa", b="bb")
    sleep(10)
    stop_daemon()
    log.info("exit")



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
jsondot is utility to make json/dict object accessible in the "." way.

##########
#Example 1: load, update, and save JSON file
##########

data = jsondot.load('path/to/my.json')
print(data.hello.message)
data.hello.message = 'Hello, mortal.'
jsondot.save(data, 'path/to/my.json')


##########
#Example 2: decorate an existing python dict
##########

my_dict = jsondot.dotify(my_dict)
print(my_dict.key1.key2)
"""


import copy
import json
import os.path
import random
import time
from typing import Any

import portalocker
import yaml


# TODO: how to remove the lib-specific dependency from this utility class?
# Register the represent_dict function with SafeDumper
def _represent_dict(dumper, data):
    return dumper.represent_dict(data.items())


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __lt__(self, other):
        # to support tuple-based sort. (dotdict_a < dotdict_b)
        return True

    def __deepcopy__(self, memo=None):
        # To workaround deepcopy with some python version
        new = {}
        for key in self.keys():
            new[key] = copy.deepcopy(self[key], memo=memo)
        return new


yaml.SafeDumper.add_representer(dotdict, _represent_dict)


def dotify(target: Any) -> Any:
    """Deeply convert an object from dict to dotdict"""

    # If already dotified, skip
    if isinstance(target, dotdict):
        return target
    if isinstance(target, list):
        for i in range(len(target)):
            target[i] = dotify(target[i])
        return target
    if isinstance(target, dict):
        for k in target:
            target[k] = dotify(target[k])
        return dotdict(target)

    # Return unchanged
    return target


def undot(target: Any) -> Any:
    """Deeply convert an object from dotdict to plain dictd"""
    if isinstance(target, list):
        for i in range(len(target)):
            target[i] = undot(target[i])
        return target
    if isinstance(target, dict):
        ret = {}
        for k in target:
            ret[k] = undot(target[k])
        return ret
    return target


def _is_primitive(obj):
    return isinstance(obj, str) or isinstance(obj, bool) or isinstance(obj, int) or isinstance(obj, float)


def plain(target: Any) -> Any:
    """Deeply convert a dotdict from dict"""
    if _is_primitive(target):
        return target

    if isinstance(target, list):
        for i in range(len(target)):
            target[i] = plain(target[i])
        return target
    if isinstance(target, dict):
        for k in target:
            target[k] = plain(target[k])
        return dict(target)


def parse(text: str) -> dotdict:
    dict = json.loads(text)
    return dotify(dict)


def _lock_with_retry(file, for_read: bool):
    if for_read:
        flags = portalocker.LockFlags.SHARED
    else:
        flags = portalocker.LockFlags.EXCLUSIVE  # | portalocker.LockFlags.NON_BLOCKING

    max_retry = 6
    retry = 0
    backoff_seconds = 0.05
    while True:
        try:
            portalocker.lock(file, flags)
            return
        except Exception as e:
            retry += 1
            if retry > max_retry:
                raise Exception(f"Fail locking file {file}") from e
            random_delay = backoff_seconds + random.uniform(0, 0.1)
            time.sleep(random_delay)
            backoff_seconds += backoff_seconds


def load(file: str, default: Any = None, lock: bool = False) -> Any:
    try:
        if not os.path.exists(file):
            return dotify(default)
        with open(file) as json_file:
            if lock:
                _lock_with_retry(json_file, True)
            dict = json.load(json_file)
        return dotify(dict)
    except Exception as e:
        raise Exception(f"Error loading file {file}") from e


def save(data: dict, file, format=True, lock: bool = False) -> None:
    with open(file, "w") as outfile:
        if lock:
            _lock_with_retry(outfile, False)
        if format:
            json.dump(data, outfile, indent="\t", default=vars)
        else:
            json.dump(data, outfile)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import config as config
from . import profile as profile
from . import state as state
from . import var_template as var_template
from ._init import init as init
from ._init import init_cli as init_cli
from .profile_store import profile_store as profile_store
from .util import CtxpException as CtxpException
from .util import choose as choose
from .util import error as error
from .util import launch_text_editor as launch_text_editor
from .util import panic as panic



from . import context
from .util import CtxpException

_recent = dict(context.get(".recent", default={}))


def set(k: str, v):
    _recent[k] = v
    context.set(".recent", _recent)
    return v


def unset(k: str):
    ret = None
    if k in _recent:
        ret = _recent.pop(k)
    context.set(".recent", _recent)
    return ret


def unset_all():
    _recent.clear()
    context.set(".recent", _recent)


def get(k: str):
    return _recent.get(k)


def all():
    return _recent


def require(provided, k: str):
    if not k:
        raise Exception("Missing 'key' in recent.require(current, key).")

    last = _recent.get(k)
    if provided:
        if last != provided:
            _recent[k] = provided
            context.set(".recent", _recent)
        return provided
    if not last:
        raise CtxpException("Missing parameter: " + k)
    return last


class Bag:
    def __init__(self, name: str):
        self._name = name
        self._changed = False

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self._changed:
            context.set(".recent", _recent)

    def get(self):
        return _recent.get(self._name)

    def set(self, v):
        if v != _recent.get(self._name):
            _recent[self._name] = v
            self._changed = True
        return v

    def unset(self):
        if self._name in _recent:
            self._changed = True
            return _recent.pop(self._name)


def of(k: str):
    return Bag(k)


class helper:
    @staticmethod
    def default_list(array: list, name: str):
        with of(name) as r:
            if len(array) == 1:
                r.set(array[0]["id"])
            else:
                r.unset()
